/**
    FitsIO.c - Input/Output Control for reading and writing images

    Copyright 1992 Barry Solomon

    Permission to use, copy, modify, distribute this software is hereby
    granted without fee, provided that this Copyright notice appear in all
    copies.  It is provided "as is" without express or implied warranty.

    Revisions
    01/21/92	BDS	Created.
    02/18/92	BDS	Modified to read multiple bitpix formats
**/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "FitsImageDefs.h"
#include "FitsDefs.h"
#include "FitsIODefs.h"
#include "FitsMenuConstants.h"

/************************************************************* power */
double power( z, n)
/**
   Calculate z to the nth power
**/
    int z, n;
{
    double a, r;
/*-------------------------------------------------------------------*/
    a = z;
    r = 1;

    if (n<0) {
        r = power( z, -n);
        return (1/r);
    }
    else {
        while (n>0) {
            if (n & 1)
                r = r * a;
            n = n>>1; /* n\2 */
            a *= a;
        }
        return (r);
    }


} /* end power() */

/***************************************************** strtod_by_me */
double strtod_by_me( string)
/**
    Translate a string into a double precision number
**/
    char *string;
{
    int exp = 0;
    double  result, sign, exp_sign, q;
    char    ch, *ch_ptr;

#define gch {         \
      ch = *ch_ptr++; \
      if (ch == 0) {  \
        ch_ptr--;     \
        ch = -1;      \
      }               \
    }
/*------------------------------------------------------------------*/
    exp    = 0;
    q      = 10;
    result = 0.0;
    sign   = 1.0;
    exp_sign   = 1.0;

    ch_ptr = string;

    gch;

    while (ch == ' ') {gch} /* skip any blanks */

    if (ch == '-') {
        sign = -1.0;
        gch;
    }
    if (ch == '+') {
        gch;
    }

    while (isdigit(ch)) {
        result = (result*10.0) + (ch - '0');
        gch;
    }

    if (ch == '.') {
        gch;
        while (isdigit(ch)) {
            result = (result + (ch - '0')/q);
            q *= 10;
            gch;
        }
    }

    if (tolower(ch) == 'e') {

        gch;
        if (ch == '-') {
            exp_sign = -1.0;
            gch;
        }
        if (ch == '+') {
            gch;
        }

        while (isdigit(ch)) {
            exp = (exp * 10) + (ch - '0');
            gch;
        }
        exp *= exp_sign;

    }

    result *= power( 10, exp);

    return (sign * result);

} /* end strtod_by_me() */

/***************************************************** keyword_to_keycode */
int keyword_to_keycode( keyword, keycode)
/**
    Parse the keyword text to an integer code
**/
    char *keyword;
    keycode_t  *keycode;
{
    int  index=0;
    int  xlated=FALSE;
/*----------------------------------------------------------------------- */
    *keycode = UNKNOWN;

    /* Loop through the keycodes for a matching keyword
       When a match is found, keycode's value is set to keycode[index].cd.
    */
    while (!xlated) {

        if (!strncmp( keyword, keycodes[index].nm, KEYWORDLEN)) {
            xlated   = TRUE;
            *keycode = keycodes[index].cd;
        }
        else
            xlated = (!strncmp(keycodes[index].nm, "END", 3));

        index++;

    } /* end while(!xlated) */

    return( xlated);

} /* end keyword_to_keycode() */

/************************************************* read_image_headers */
int  read_image_headers( FitsFilePtr, FImagePtr)
/**
    Read all the cards for the current record

    FITS header card
        Keyword is left justified in columns 1-8
           '=' - in columns 9-10 (except comment cards)
        Parameter value - in ANSI FORTRAN 77 list
               - directed read format anywhere in columns 11 - 80
           '/' - optional, terminates value field
        Comments - follows the /
**/
    FILE	*FitsFilePtr;
    fimage_t	*FImagePtr;
{
    keycode_t   keycode;
    int         card_number = 0;
    char       *c_ptr, comment[80];
    int        done = FALSE, hasequ;
    char        current_card[81];
/*--------------------------------------------------------------------*/

    while (!done) {
/*
 *      Read the keycode to figure out how to translate the rest of the card
 */
	fgets(current_card, 81, FitsFilePtr);
	keyword_to_keycode(current_card, &keycode);

	if (ferror(FitsFilePtr) || (card_number++ > MAXCARD)) {
	    printf("Error occured during read operation\n");
	    return (READ_ERROR);
	}

/*
 *      Remove equal sign and any blanks before value
 */
	c_ptr = current_card;

	while ((*c_ptr != '=') && (*c_ptr != '\0'))
	    c_ptr++;
	hasequ = (c_ptr != 0);

	if (hasequ)
	    while ((*(++c_ptr) == ' ') && (*c_ptr != 0));

	switch (keycode) {

	    case SIMPLE:
		simple = *c_ptr++;
		break;

	    case BITPIX:
		FImagePtr->bitpix = strtol(c_ptr, comment, 10);
		break;

	    case NAXIS:
		naxis = strtol(c_ptr, comment, 10);
		break;

	    case NAXIS1:
		FImagePtr->naxis1 = strtol(c_ptr, comment, 10);
		break;

	    case NAXIS2:
		FImagePtr->naxis2 = strtol(c_ptr, comment, 10);
		break;

	    case BZERO:
		FImagePtr->bzero = strtod_by_me(c_ptr, comment);
		break;

	    case BSCALE:
		FImagePtr->bscale = strtod_by_me(c_ptr, comment);
		break;

	    case CORRUPTED_FILE:
		printf("\nProblems encountered during read.\n");
		fclose(FitsFilePtr);
		return (READ_ERROR);

	    case END:
		/* Read the remaining cards of this record */
		while (card_number++ % 36)
		    fgets(current_card, 81, FitsFilePtr);
		done = TRUE;
		break;

	    default:
		break;

	}			/* end switch(keyword) */

    }				/* end while(!done) */

    return (SUCCESSFUL);

}				/* read_image_headers() */

/****************************************************** read_image_data */
int  read_image_data( FitsFilePtr, FImagePtr)
/**
    Read the FITS data and store into a fits image structure
**/
    FILE	*FitsFilePtr;
    fimage_t	*FImagePtr;
{
    int             Count, NumBytes, dataread;
    char           *RawData;
    float           MinVal = 0, MaxVal = 0;
    int             num_pixels, index;
    float          *F1Data;
    char           *RawDataPtr8;
    short          *RawDataPtr16;
    long           *RawDataPtr32;
    double         *RawDataPtr64;
/*--------------------------------------------------------------------*/

    Count = FImagePtr->naxis1 * FImagePtr->naxis2;

    switch (FImagePtr->bitpix) {
    case 64:
	(double *) RawData = (double *) malloc(Count * sizeof(double));
	break;
    case 32:
	(long *) RawData = (long *) malloc(Count * sizeof(long));
	break;
    case 16:
	(short *) RawData = (short *) malloc(Count * sizeof(short));
	break;
    case 8:
	(char *) RawData = (char *) malloc(Count * sizeof(char));
	break;
    }

    FImagePtr->data_max = 0;
    FImagePtr->data_min = 0;

    NumBytes = FImagePtr->bitpix / 8;
    dataread = fread(RawData, NumBytes, Count, FitsFilePtr);

    if (dataread != Count) {
	printf("data read error\n");
	return (READ_ERROR);
    }

    num_pixels = FImagePtr->naxis1 * FImagePtr->naxis2;
    FImagePtr->fdata = (float *) malloc(num_pixels * sizeof(float));
    F1Data = FImagePtr->fdata;

    RawDataPtr8  = (char *) RawData;
    RawDataPtr16 = (short *) RawData;
    RawDataPtr32 = (long *) RawData;
    RawDataPtr64 = (double *) RawData;

/*
 *  Initial translation:  REAL = DATA * BSCALE + BZERO
 */
    switch (FImagePtr->bitpix) {
    case 64:
	for (index = 0; index < num_pixels; index++) {
	    *F1Data = *RawDataPtr64 * FImagePtr->bscale + FImagePtr->bzero;
	    MinVal = (*RawDataPtr64 < MinVal) ? *RawDataPtr64 : MinVal;
	    MaxVal = (*RawDataPtr64 > MaxVal) ? *RawDataPtr64 : MaxVal;
	    RawDataPtr64++;
	    F1Data++;
	}
	break;
    case 32:
	for (index = 0; index < num_pixels; index++) {
	    *F1Data = *RawDataPtr32 * FImagePtr->bscale + FImagePtr->bzero;
	    MinVal = (*RawDataPtr32 < MinVal) ? *RawDataPtr32 : MinVal;
	    MaxVal = (*RawDataPtr32 > MaxVal) ? *RawDataPtr32 : MaxVal;
	    RawDataPtr32++;
	    F1Data++;
	}
	break;
    case 16:
	for (index = 0; index < num_pixels; index++) {
	    *F1Data = *RawDataPtr16 * FImagePtr->bscale + FImagePtr->bzero;
	    MinVal = (*RawDataPtr16 < MinVal) ? *RawDataPtr16 : MinVal;
	    MaxVal = (*RawDataPtr16 > MaxVal) ? *RawDataPtr16 : MaxVal;
	    RawDataPtr16++;
	    F1Data++;
	}
	break;
    case 8:
	for (index = 0; index < num_pixels; index++) {
	    *F1Data = *RawDataPtr16 * FImagePtr->bscale + FImagePtr->bzero;
	    MinVal = (*RawDataPtr16 < MinVal) ? *RawDataPtr8 : MinVal;
	    MaxVal = (*RawDataPtr16 > MaxVal) ? *RawDataPtr8 : MaxVal;
	    RawDataPtr8++;
	    F1Data++;
	}
	break;
    }


    FImagePtr->data_min = MinVal * FImagePtr->bscale + FImagePtr->bzero;
    FImagePtr->data_max = MaxVal * FImagePtr->bscale + FImagePtr->bzero;
    FImagePtr->image_min = FImagePtr->data_min;
    FImagePtr->image_max = FImagePtr->data_max;

/*
 *  Transform the image values to lie within the range 0.0 to 1.0
 */
    F1Data = FImagePtr->fdata;
    MinVal = FImagePtr->data_min;
    MaxVal = FImagePtr->data_max - MinVal;
    for (index = 0; index < num_pixels; index++) {
	*F1Data = (*F1Data - MinVal) / MaxVal;
	F1Data++;
    }

    free(RawData);

    return (SUCCESSFUL);

}/* end read_image_data() */
